#include <Arduino.h>
// We use the TiCoServo library because it works with the NeoPixel library
#include <Adafruit_TiCoServo.h>
#include <Adafruit_NeoPixel.h>
#include <AccelStepper.h>
#include <LiquidCrystal_I2C.h>
#include <MultiStepper.h>

namespace claw
{
  // PWM port, must be a specific one to support TiCoServo
  constexpr int port = 9; // aka X+, X-
  // Max and min pulse width in microseconds
  // Though we don't actually need to put these explicitly
  constexpr int minUs = 544;
  constexpr int maxUs = 2400;

  // Different positions for the claw
  constexpr int pushyDeg = 84;
  constexpr int closeDeg = 78;
  constexpr int openDeg = 60;

  Adafruit_TiCoServo myservo;

  void set(const int pos) { myservo.write(pos); }

  void pushy() { set(pushyDeg); }
  void closed() { set(closeDeg); }
  void open() { set(openDeg); }

  void setup()
  {
    myservo.attach(port, minUs, maxUs);
    open();
  }
}

namespace joystick
{
  constexpr int portX = 0; // aka A0, Abort
  constexpr int portY = 1; // aka A1, Hold
  constexpr int portSW = 13; // aka D13, SpnDir

  void setup()
  {
    // note: analog read resolution 10 bit is default on arduinos
    // so values range from 0-1023 from analogRead
    pinMode(portX, INPUT);
    pinMode(portY, INPUT);
    pinMode(portSW, INPUT_PULLUP);
  }

  // returns joystick X -512 (left)to 511 (right)
  int getX()
  {
    const int val = analogRead(portX) - 512;
    if (abs(val) < 90) return 0; // deadband to prevent drift
    return -val;
  }

  // returns joystick Y -512 (down)to 511 (up)
  int getY()
  {
    const int val = analogRead(portY) - 512;
    if (abs(val) < 90) return 0; // deadband to prevent drift
    return val;
  }

  // returns whether the switch on the joystick is pressed
  // this is currently unused
  bool getSwitch() { return !digitalRead(portSW); }
}

namespace button
{
  constexpr int buttonLed1 = 11; // z+/z-
  constexpr int buttonSw1 = 10; // y+/y-

  void setup()
  {
    pinMode(buttonLed1, OUTPUT);
    pinMode(buttonSw1, INPUT_PULLUP);
  }

  void setLed(const int val) { digitalWrite(buttonLed1, val); }

  void setLedAnalog(const int val) { analogWrite(buttonLed1, val); }

  void ledOn() { setLed(HIGH); }
  void ledOff() { setLed(LOW); }

  bool readSwitch()
  {
    // we need to invert it because the button connects to ground when it is pressed
    // and we are using INPUT_PULLUP so it is HIGH when not pressed
    return !digitalRead(buttonSw1);
  }
}

namespace neopixel
{
  constexpr int numpixels = 52; // wow this is a lot of neopixels!
  constexpr int pin = 12; // SpnEn
  Adafruit_NeoPixel pixels(numpixels, pin, NEO_GRB + NEO_KHZ800);

  void setup()
  {
    //pixels.setBrightness(128);
    pixels.begin();
  }

  void patternBlueSnakes()
  {
    pixels.clear();
    const unsigned long offset = millis() / 80;
    for (int i = 0; i < numpixels; i++)
    {
      const unsigned long thing = ((i + offset) / 7) & 1;
      if (thing) { pixels.setPixelColor(i, Adafruit_NeoPixel::Color(0, 0, 160)); }
    }
    pixels.show();
  }

  void patternRedSnakes()
  {
    pixels.clear();
    const unsigned long int offset = millis() / 80;
    for (int i = 0; i < numpixels; i++)
    {
      const unsigned long int thing = ((i + offset) / 7) & 1;
      if (thing) { pixels.setPixelColor(i, Adafruit_NeoPixel::Color(160, 0, 0)); }
    }
    pixels.show();
  }

  void patternSweepyGreen()
  {
    pixels.clear();
    for (int i = 0; i < numpixels; i++) { pixels.setPixelColor(i, Adafruit_NeoPixel::Color(0, (i * 17) % 250 + 5, 0)); }
    pixels.show();
  }

  void clear()
  {
    pixels.clear();
    pixels.show();
  }
}

namespace lcd
{
  // uses I2C port
  LiquidCrystal_I2C lcd(0x27, 16, 2);

  byte claw[8] = {
    B00100,
    B00100,
    B00100,
    B11111,
    B10001,
    B10001,
    B11011,
  };
  byte joystick[8] = {
    B11111,
    B01110,
    B01110,
    B00100,
    B11111,
    B11111,
    B11111,
  };

  byte button[8] = {
    B00000,
    B01110,
    B01110,
    B11111,
    B11111,
    B11111,
    B11111,
  };

  void displayStart()
  {
    lcd.clear();
    lcd.setCursor(1, 0);
    lcd.print("\7 Claw Machine");
    lcd.setCursor(2, 1);
    lcd.print("\5 Start");
  }

  void displayStarting()
  {
    lcd.clear();
    lcd.setCursor(1, 0);
    lcd.print("\7 Claw Machine");
    lcd.setCursor(2, 1);
    lcd.print("Starting!");
  }

  void displayMain()
  {
    lcd.clear();
    lcd.setCursor(1, 0);
    lcd.print("\7 Claw Machine");
    lcd.setCursor(0, 1);
    lcd.print("\6Move  \5Use Claw");
  }

  void displayMoving()
  {
    lcd.clear();
    lcd.setCursor(1, 0);
    lcd.print("\7 Claw Machine");
    lcd.setCursor(0, 1);
    lcd.print("\7 Picking up.. \7");
  }

  void setup()
  {
    lcd.init();
    lcd.createChar(5, button);
    lcd.createChar(6, joystick);
    lcd.createChar(7, claw);
    lcd.backlight();
  }
}

namespace xyaxis
{
  // pins
  constexpr int stepA = 2; // built in to cnc shield
  constexpr int dirA = 5; // built in to cnc shield
  constexpr int stepB = 3; // built in to cnc shield
  constexpr int dirB = 6; // built in to cnc shield

  constexpr float maxMotorSpeed = 500;
  constexpr float maxMotorAccel = 1200;
  constexpr int motorHome = 0;

  AccelStepper stepperA(AccelStepper::DRIVER, stepA, dirA);
  AccelStepper stepperB(AccelStepper::DRIVER, stepB, dirB);

  MultiStepper group;

  void setup()
  {
    // note: acceleration isn't actually used since we are using setSpeed
    // except when we call home()
    stepperA.setAcceleration(maxMotorAccel);
    stepperA.setMaxSpeed(maxMotorSpeed);
    stepperB.setAcceleration(maxMotorAccel);
    stepperB.setMaxSpeed(maxMotorSpeed);

    group.addStepper(stepperA);
    group.addStepper(stepperB);
  }

  void home()
  {
    long positions[2];
    positions[0] = motorHome;
    positions[1] = motorHome;
    group.moveTo(positions);
    group.runSpeedToPosition();
  }

  // speedX -maxMotorSpeed to maxMotorSpeed
  // speedY -maxMotorSpeed to maxMotorSpeed
  // note that: Y+ is away from front, X+ is towards right
  void setSpeeds(const float speedX, const float speedY)
  {
    float speedA = speedX + speedY;
    float speedB = speedX - speedY;
    // note: max power = maxMotorSpeed
    // we need to constrain each to max power
    // since they could have gone greater (maxMotorSpeed + maxMotorSpeed = 2*maxMotorSpeed, etc)
    // and we want to keep the same ratio between speedA and speedB
    float speedAProportion = speedA / maxMotorSpeed;
    float speedBProportion = speedB / maxMotorSpeed;
    speedAProportion = abs(speedAProportion);
    speedBProportion = abs(speedBProportion);

    const float dividend = max(max(speedAProportion, speedBProportion), 1);
    // divide the things
    speedA /= dividend;
    speedB /= dividend;

    // flip signs to make it go the correct direction
    stepperA.setSpeed(-speedA);
    stepperB.setSpeed(-speedB);

    stepperA.runSpeed();
    stepperB.runSpeed();
  }
}

namespace zaxis
{
  constexpr int stepZ = 4; // built in to cnc shield
  constexpr int dirZ = 7; // built in to cnc shield

  constexpr float maxMotorAccel = 1200;
  constexpr float maxMotorSpeed = 800;

  constexpr int motorHome = 0;
  constexpr int motorDown = -10000;

  AccelStepper stepperZ(AccelStepper::DRIVER, stepZ, dirZ);

  void setup()
  {
    stepperZ.setAcceleration(maxMotorAccel);
    stepperZ.setMaxSpeed(maxMotorSpeed);
  }

  // note: these methods block!!

  void home() { stepperZ.runToNewPosition(motorHome); }
  void down() { stepperZ.runToNewPosition(motorDown); }
}

void setup()
{
  claw::setup();
  joystick::setup();
  button::setup();
  xyaxis::setup();
  zaxis::setup();
  neopixel::setup();
  button::setup();
  lcd::setup();
}

void grabTheThing()
{
  zaxis::down();
  delay(500);
  claw::closed();
  delay(1000);
  claw::pushy();
  delay(1000);
  zaxis::home();
  delay(500);
  xyaxis::home();
  delay(500);
  claw::open();
  delay(500);
}

void runningJoystickLoop()
{
  float joystickX = joystick::getX();
  float joystickY = joystick::getY();
  constexpr float scale = xyaxis::maxMotorSpeed / 512.0; // scale from the joystick to the motor speeds
  joystickX *= scale;
  joystickY *= scale;
  xyaxis::setSpeeds(joystickX, joystickY);
}

void loop()
{
  // pre-starting state
  lcd::displayStart();
  while (!button::readSwitch())
  {
    button::setLedAnalog((millis() / 4) % 255); // blink leds
    neopixel::patternRedSnakes();
  }
  neopixel::clear();
  button::ledOn();
  lcd::displayStarting();
  delay(1500);
  // moving joystick state
  lcd::displayMain();
  while (!button::readSwitch())
  {
    runningJoystickLoop(); // run the loop
    neopixel::patternBlueSnakes();
  }
  // moving claw state
  button::ledOff();
  lcd::displayMoving();
  neopixel::clear();
  delay(1500);
  neopixel::patternSweepyGreen();
  grabTheThing();
  neopixel::clear();
  delay(1500);
  neopixel::clear();
}
